@@ -1,16 +1,19 @@
 """nprintML command-line interface entry-point"""
 import argparse
 import itertools
+import multiprocessing
 import os
 import pathlib
 import re
 import sys
 import time
+import toml
 
 import argparse_formatter
 
 import nprintml
 from nprintml.pipeline import Pipeline
+from nprintml.util import NumericRangeType
 
 # ensure default steps of pipeline auto-load
 import nprintml.net.step
@@ -26,23 +29,28 @@ ANIMALS = ('aardvark', 'bison', 'canary', 'dalmation', 'emu', 'falcon', 'gnu',
 
 def execute(argv=None, **parser_kwargs):
     """Execute the nprintml CLI command."""
+    args = None
+
     try:
         parser = build_parser(**parser_kwargs)
 
         pipeline = Pipeline(parser)
 
-        args = parser.parse_args(argv)
+        args = parser.parse_args(argv, Namespace())
 
         check_output_directory(args)
 
-        for (step, results) in pipeline(args):
+        for (step, results) in pipeline(parser, args, meta={}):
             print(step, results, sep=' → ')
 
+        with args.meta_file.open('w') as meta_file:
+            dump_meta(pipeline.results, meta_file)
+
         print('done →', pipeline.results)
     except KeyboardInterrupt:
         print('interrupted ✕')
     except Exception as exc:
-        if args.traceback:
+        if args is not None and args.traceback:
             raise
 
         print(f'error:{exc_repr(exc)} ✕')
@@ -111,14 +119,22 @@ def build_parser(**parser_kwargs):
         help="print exception tracebacks",
     )
 
-    cpu_available_count = len(os.sched_getaffinity(0))
+    try:
+        # glibc-only
+        sched_getaffinity = os.sched_getaffinity
+    except AttributeError:
+        # Note: this *may* be inaccurate in some shared environs
+        cpu_available_count = multiprocessing.cpu_count()
+    else:
+        cpu_available_count = len(sched_getaffinity(0))
+
     parser.add_argument(
         '--concurrency',
         default=cpu_available_count,
         metavar='INTEGER',
-        type=int,
-        help="maximum number of concurrent processes to use (defaults to number "
-             f"reported by scheduler: {cpu_available_count})",
+        type=NumericRangeType(int, (0, None)),
+        help="maximum number of concurrent processes to apply to data preparation "
+             f"(defaults to number reported by scheduler: {cpu_available_count})",
     )
 
     output_default = get_default_directory()
@@ -140,6 +156,15 @@ def build_parser(**parser_kwargs):
     return parser
 
 
+class Namespace(argparse.Namespace):
+
+    meta_file_name = 'meta.toml'
+
+    @property
+    def meta_file(self):
+        return self.outdir / self.meta_file_name
+
+
 def exc_repr(exc):
     """Construct representation of given exception appropriate for
     printed output.
@@ -202,3 +227,15 @@ def check_output_directory(args):
             args.__parser__.error(f'output path exists and is not a directory: {args.outdir}')
     else:
         args.outdir.mkdir(parents=True)
+
+
+def dump_meta(results, meta_file):
+    step_timing = (
+        (step.__name__, tuple(timing))
+        for (step, timing) in results.__timing_steps__.items()
+    )
+    meta_timing = dict(step_timing, total=tuple(results.__timing__))
+
+    meta = dict(results.meta, timing=meta_timing)
+
+    toml.dump(meta, meta_file)
