@@ -28,8 +28,9 @@ class Label(pipeline.Step):
     Returns a `LabelResult`.
 
     """
+    __pre_provides__ = ('labels',)
     __provides__ = LabelResult
-    __requires__ = ('nprint_path',)
+    __requires__ = ('nprint_stream',)
 
     def __init__(self, parser):
         group_parser = parser.add_argument_group(
@@ -67,19 +68,28 @@ class Label(pipeline.Step):
             help="drop columns which do not appear to provide any predictive signal",
         )
         group_parser.add_argument(
-            '--no-write-features',
+            '--no-save-features',
             action='store_false',
-            default=True,
-            dest='write_features',
+            dest='save_features',
             help="disable writing of features to disk (enabled by default for inspection & reuse)",
         )
 
-    def __call__(self, args, results):
-        aggregator_class = aggregators[args.aggregator]
-        aggregator = aggregator_class(results.nprint_path, args.label_file)
-        features = aggregator(compress=args.compress, sample_size=args.sample_size)
+        self.aggregator = None
 
-        if args.write_features:
+    def __pre__(self, parser, args, results):
+        aggregator_class = aggregators[args.aggregator]
+        self.aggregator = aggregator_class(args.label_file)
+
+        results.labels = self.aggregator.labels
+
+    def __call__(self, args, results):
+        features = self.aggregator(
+            results.nprint_stream,
+            compress=args.compress,
+            sample_size=args.sample_size,
+        )
+
+        if args.save_features:
             outdir = args.outdir / 'feature'
             outdir.mkdir()
 
