@@ -7,20 +7,24 @@ Graphs may be constructed (with sane defaults) to visually illustrate
 classifier performance.
 
 """
+import csv
 import itertools
 
 import matplotlib.pyplot as plt
 import seaborn as sns
-from autogluon import TabularPrediction as task
+from autogluon.tabular import TabularPredictor
 from sklearn.model_selection import train_test_split
 from sklearn.preprocessing import LabelBinarizer
 from sklearn.metrics import (
-    roc_curve,
     auc,
-    precision_recall_curve,
     average_precision_score,
+    balanced_accuracy_score,
     confusion_matrix,
     ConfusionMatrixDisplay,
+    f1_score,
+    precision_recall_curve,
+    roc_auc_score,
+    roc_curve,
 )
 
 
@@ -57,11 +61,9 @@ class AutoML:
         'best_quality',
     )
 
-    N_THREADS = None
-
     TEST_SIZE = 0.3
 
-    TIME_LIMITS = 5 * 60
+    TIME_LIMIT = 5 * 60
 
     VERBOSITY = 1
 
@@ -78,26 +80,30 @@ class AutoML:
         return self.outpath / self.models_dirname
 
     def __call__(self, test_size=TEST_SIZE, eval_metric=EVAL_METRIC, quality=QUALITY,
-                 time_limits=TIME_LIMITS, n_threads=N_THREADS, verbosity=VERBOSITY):
+                 time_limit=TIME_LIMIT, verbosity=VERBOSITY):
         """Train, test, and evaluate models."""
         (train_data, test_data) = train_test_split(self.data, test_size=test_size)
-        predictor = self.train(train_data, eval_metric, quality, time_limits,
-                               n_threads, verbosity=verbosity)
+        predictor = self.train(train_data, eval_metric, quality, time_limit,
+                               verbosity=verbosity)
         self.test(predictor, test_data)
         self.graph_all(predictor, test_data)
 
     def train(self, train_data, eval_metric=EVAL_METRIC, quality=QUALITY,
-              time_limits=TIME_LIMITS, n_threads=N_THREADS, verbosity=VERBOSITY):
+              time_limit=TIME_LIMIT, verbosity=VERBOSITY):
         """Train prospective models."""
         # predictor gives us default access to the *best* predictor that
         # was trained on the task (otherwise we're just wrapping AutoGluon)
-        return task.fit(train_data=train_data, label='label',
-                        eval_metric=eval_metric,
-                        output_directory=self.outpath,
-                        time_limits=time_limits,
-                        presets=self.QUALITY_PRESETS[quality],
-                        nthreads_per_trial=n_threads,
-                        verbosity=verbosity)
+        task = TabularPredictor(
+            label='label',
+            eval_metric=eval_metric,
+            path=self.outpath,
+            verbosity=verbosity,
+        )
+        return task.fit(
+            train_data=train_data,
+            time_limit=time_limit,
+            presets=self.QUALITY_PRESETS[quality],
+        )
 
     def test(self, predictor, test_data):
         """Evaluate models on the test set and write the results to file."""
@@ -118,10 +124,28 @@ class AutoML:
         binarized_labels = binarizer.transform(y_true)
 
         y_pred = predictor.predict(test_no_label)
-        y_proba = predictor.predict_proba(test_no_label)
+        y_proba = predictor.predict_proba(test_no_label, as_multiclass=False)
         self.make_pr(binarizer.classes_, binarized_labels, y_proba)
         self.make_roc(binarizer.classes_, binarized_labels, y_proba)
         self.make_cfmx(binarizer.classes_, y_true, y_pred)
+        self.make_stat_report(binarizer.classes_, binarized_labels, y_true, y_pred, y_proba)
+
+    def make_stat_report(self, classes, binarized_labels, y_true, y_pred, y_proba):
+        """Write report of basic statistics that users might put into a results summary table."""
+        ba = balanced_accuracy_score(y_true, y_pred)
+        f1_macro = f1_score(y_true, y_pred, average='macro')
+        f1_micro = f1_score(y_true, y_pred, average='micro')
+        multi_class = 'raise' if len(classes) == 2 else 'ovr'
+        roc_macro = roc_auc_score(y_true, y_proba, average='macro', multi_class=multi_class)
+        roc_weighted = roc_auc_score(y_true, y_proba, average='weighted', multi_class=multi_class)
+
+        stats_path = self.outpath / 'stat-report.csv'
+        with stats_path.open('w') as fd:
+            writer = csv.writer(fd)
+            writer.writerow(('f1_micro', 'f1_macro', 'balanced_accuracy',
+                             'roc_macro', 'roc_weighted'))
+            writer.writerow(f'{value:.2f}' for value in (f1_micro, f1_macro, ba,
+                                                         roc_macro, roc_weighted))
 
     def make_cfmx(self, classes, y_true, y_pred):
         """Make confusion matrix without printing exact values.
@@ -150,7 +174,7 @@ class AutoML:
             plt_ax = self._make_binary_pr(y_true_bin.ravel(), y_proba.ravel())
         else:
             for i, class_label in enumerate(classes):
-                plt_ax = self._make_binary_pr(y_true_bin[:, i], y_proba[:, i],
+                plt_ax = self._make_binary_pr(y_true_bin[:, i], y_proba.iloc[:, i],
                                               class_label=class_label)
 
         # Split up the line styles to make them unique
@@ -188,7 +212,7 @@ class AutoML:
             plt_ax = self._make_binary_roc(y_true_bin.ravel(), y_proba.ravel())
         else:
             for i, class_label in enumerate(classes):
-                plt_ax = self._make_binary_roc(y_true_bin[:, i], y_proba[:, i],
+                plt_ax = self._make_binary_roc(y_true_bin[:, i], y_proba.iloc[:, i],
                                                class_label=class_label)
 
         # Split up the line styles to make them unique
