@@ -24,15 +24,9 @@ def nprint(*args, stdin=None, stdout=None, stderr=None, check=True):
     Returns the resulting `subprocess.CompletedProcess`.
 
     """
-    # ensure we know what we're running via which
-    cmd = shutil.which('nprint')
-
-    if cmd is None:
-        raise NoCommand
-
     try:
         return subprocess.run(
-            (cmd,) + args,
+            _nprint_args(*args),
             stdin=stdin,
             stdout=stdout,
             stderr=stderr,
@@ -42,6 +36,44 @@ def nprint(*args, stdin=None, stdout=None, stderr=None, check=True):
         raise CommandError from exc
 
 
+def nPrintProcess(*args, stdin=None, stdout=None, stderr=None):
+    """Construct a subprocess to execute the `nprint` command.
+
+    Argumentation reflects that of `subprocess.Popen`.
+
+    The `PIPE` sentinel is made available via the object of the
+    constructor:
+
+        nPrintProcess.PIPE
+
+    If the `nprint` command cannot be found on the PATH, then
+    `NoCommand` is raised. This exception is also made available on the
+    constructor --
+
+        nPrintProcess.NoCommand
+
+    Returns the constructed `subprocess.Popen` object with which to
+    monitor the running command.
+
+    """
+    return subprocess.Popen(
+        _nprint_args(*args),
+        stdin=stdin,
+        stdout=stdout,
+        stderr=stderr,
+    )
+
+
+def _nprint_args(*args):
+    # ensure we know what we're running via which
+    cmd = shutil.which('nprint')
+
+    if cmd is None:
+        raise NoCommand
+
+    return (cmd,) + args
+
+
 class nPrintError(Exception):
     pass
 
@@ -57,3 +89,6 @@ class NoCommand(nPrintError):
 nprint.CommandError = CommandError
 nprint.NoCommand = NoCommand
 nprint.PIPE = subprocess.PIPE
+
+nPrintProcess.NoCommand = NoCommand
+nPrintProcess.PIPE = subprocess.PIPE
