@@ -1,24 +1,33 @@
 """Pipeline Step to extract networking traffic via nPrint: Net"""
 import argparse
+import collections
+import concurrent.futures
 import itertools
 import os
 import pathlib
 import re
 import sys
 import textwrap
-import time
 import typing
 
 import nprintml
 from nprintml import pipeline
-from nprintml.util import HelpAction
+from nprintml.util import (
+    HelpAction,
+    NamedBytesIO,
+    prime_iterator,
+)
 
-from .execute import CommandError, nprint, nPrintProcess
+from . import execute
+
+
+NPRINT_QUEUE_MAX = 10 ** 6
 
 
 class NetResult(typing.NamedTuple):
     """Pipeline Step results for Net"""
     nprint_path: pathlib.Path
+    nprint_stream: typing.Generator[NamedBytesIO, None, None]
 
 
 class Net(pipeline.Step):
@@ -29,14 +38,20 @@ class Net(pipeline.Step):
 
     """
     __provides__ = NetResult
+    __requires__ = ('labels',)
 
     def __init__(self, parser):
+        self.args = None  # set by __pre__
+
         self.group_parser = parser.add_argument_group(
             "extraction of features from network traffic via nPrint",
 
             "(full information can be found at https://nprint.github.io/nprint/)"
         )
 
+        # record of subset of arguments NOT passed to nprint
+        self.own_arguments = set()
+
         self.group_parser.add_argument(
             '-4', '--ipv4',
             action='store_true',
@@ -123,6 +138,7 @@ class Net(pipeline.Step):
             type=FileAccessType(os.R_OK),
             help="pcap infile",
         )
+        self.own_arguments.add('pcap_file')
         self.group_parser.add_argument(
             '--pcap-dir', '--pcap_dir',
             default=(),
@@ -131,6 +147,7 @@ class Net(pipeline.Step):
             type=DirectoryAccessType(ext='.pcap'),
             help="directory containing pcap infile(s) with file extension '.pcap'",
         )
+        self.own_arguments.add('pcap_dir')
         self.group_parser.add_argument(
             '-R', '--relative-timestamps', '--relative_timestamps',
             action='store_true',
@@ -155,58 +172,68 @@ class Net(pipeline.Step):
         self.group_parser.add_argument(
             '--help-nprint-filter', '--nprint-filter-help', '--nprint_filter_help',
             action=HelpAction,
-            help_action=lambda *_parser_args: nprint('--nprint_filter_help'),
+            help_action=lambda *_parser_args: execute.nprint('--nprint_filter_help'),
             help="describe regex possibilities and exit",
         )
 
-    @staticmethod
-    def get_output_directory(args):
-        return args.outdir / 'nprint'
+        self.group_parser.add_argument(
+            '--save-nprint',
+            action='store_true',
+            help="save nPrint output(s) to disk (default: not saved)",
+        )
+        self.own_arguments.add('save_nprint')
+
+    output_tag = 'nprint'
+    default_output_name = 'netcap.npt'
+
+    @property
+    def output_directory(self):
+        return self.args.outdir / self.output_tag
+
+    def ensure_output_directory(self):
+        if self.args.save_nprint:
+            self.output_directory.mkdir()
 
     @classmethod
-    def make_output_directory(cls, args):
-        outdir = cls.get_output_directory(args)
-        outdir.mkdir()
-        return outdir
+    def get_output_name(cls, pcap_output):
+        if pcap_output is None:
+            return cls.default_output_name
 
-    @staticmethod
-    def make_output_path(outdir, pcap_file, dir_basis):
-        if not pcap_file:
-            return None
+        # pcap_output expects str and re is fast
+        return cls.get_output_name.pattern.sub('.npt', pcap_output)
 
-        npt_file = pathlib.Path(pcap_file).with_suffix('.npt')
-        if dir_basis:
-            npt_file = npt_file.relative_to(dir_basis)
-        else:
-            npt_file = npt_file.name
+    get_output_name.__func__.pattern = re.compile(r'\.pcap$', re.I)
 
-        npt_path = outdir / npt_file
+    def make_output_path(self, npt_file):
+        npt_path = self.output_directory / npt_file
 
         if npt_path.exists():
             raise FileExistsError(None, 'nPrint output path collision', str(npt_path))
 
         npt_path.parent.mkdir(parents=True, exist_ok=True)
+
         return npt_path
 
-    @staticmethod
-    def generate_files(args):
-        if args.pcap_file or args.pcap_dir:
-            # stream pair of pcap path & "basis" for reconstructing tree
-            for pcap_file in args.pcap_file:
-                yield (pcap_file, None)
+    def generate_pcaps(self):
+        if self.args.pcap_file or self.args.pcap_dir:
+            # stream pair of pcap path & short ref path for reconstructing tree
+            for pcap_file in self.args.pcap_file:
+                pcap_path = pathlib.Path(pcap_file)
+                yield (pcap_path, pcap_path.name)
 
-            for pcap_dir in args.pcap_dir:
-                for pcap_file in pcap_dir.rglob('*.pcap'):
-                    yield (pcap_file, pcap_dir)
+            for pcap_dir in self.args.pcap_dir:
+                for pcap_path in pcap_dir.rglob('*.pcap'):
+                    pcap_file = pcap_path.relative_to(pcap_dir)
+                    yield (pcap_path, str(pcap_file))
 
             return
 
         yield (None, None)
 
-    def generate_argv(self, args, pcap_file=None, npt_file=None):
+    def generate_argv(self, pcap_file=None, npt_file=None):
         """Construct arguments for `nprint` command."""
         # generate shared/global arguments
-        if args.verbosity >= 3:
+        if self.args.verbosity >= 3:
             yield '--verbose'
 
         # support arbitrary pcap infile(s)
@@ -215,11 +242,11 @@ class Net(pipeline.Step):
 
         # add group (nPrint-specific) arguments
         for action in self.group_parser._group_actions:
-            if action.dest in ('pcap_file', 'pcap_dir'):
+            if action.dest in self.own_arguments:
                 continue
 
             try:
-                value = getattr(args, action.dest)
+                value = getattr(self.args, action.dest)
             except AttributeError:
                 if action.default == argparse.SUPPRESS:
                     continue
@@ -235,56 +262,143 @@ class Net(pipeline.Step):
                     yield str(value)
 
         # add output path
-        outdir = self.get_output_directory(args)
-        outpath = npt_file or outdir / 'netcap.npt'
-        yield from ('--write_file', str(outpath))
+        if npt_file:
+            yield from ('--write_file', npt_file)
 
-    def generate_procs(self, args, pcap_files, outdir):
-        for (pcap_file, dir_basis) in pcap_files:
-            npt_file = self.make_output_path(outdir, pcap_file, dir_basis)
+    def write_meta(self, target):
+        args = ' '.join(self.generate_argv('[input_pcap]'))
+        target[self.output_tag] = {'cmd': f'nprint {args}'}
 
-            yield nPrintProcess(
-                *self.generate_argv(args, pcap_file, npt_file),
-            )
+    def filtermap_pcaps(self, pcap_files, labels=None):
+        skipped_files = collections.deque(maxlen=4)
+        skipped_count = 0
 
-    @staticmethod
-    def pool_procs(proc_stream, size, wait=None):
-        sleep_time = os.sched_rr_get_interval(0) if wait is None else wait / 1_000
+        for (pcap_path, pcap_file) in pcap_files:
+            if pcap_path:
+                if labels is not None and pcap_file not in labels.index:
+                    skipped_files.append(pcap_file)
+                    skipped_count += 1
+                    continue
 
-        pool = list(itertools.islice(proc_stream, size))
+                npt_file = self.get_output_name(pcap_file)
 
-        while any(pool):
-            if sleep_time != 0:
-                time.sleep(sleep_time)
+                yield (pcap_path, npt_file)
+            else:
+                npt_file = self.get_output_name(None)
 
-            for (index, proc) in enumerate(pool):
-                if proc and proc.poll() is not None:
-                    if proc.returncode != 0:
-                        raise CommandError
+                yield (None, npt_file)
 
-                    pool[index] = next(proc_stream, None)
+        if skipped_count > 0 and self.args.verbosity >= 1:
+            print('Skipped', skipped_count, 'PCAP file(s) missing from labels file:')
 
-    def __call__(self, args, results):
+            for skipped_file in skipped_files:
+                print(f'\t{skipped_file}')
+
+            if skipped_count > len(skipped_files):
+                print('\t...')
+
+    def execute_nprint(self, pcap_file, npt_file):
+        result = execute.nprint(
+            *self.generate_argv(pcap_file),
+            stdout=execute.nprint.PIPE,
+        )
+
+        if self.args.save_nprint:
+            out_path = self.make_output_path(npt_file)
+            out_path.write_bytes(result.stdout)
+
+        return NamedBytesIO(result.stdout, name=npt_file)
+
+    def generate_npts(self, file_stream, timing):
+        # Spawn thread pool of same size as number of concurrent
+        # subprocesses we would like to maintain.
+        with timing, concurrent.futures.ThreadPoolExecutor(
+            max_workers=self.args.concurrency,
+            thread_name_prefix=__name__,
+        ) as executor:
+            # Enqueue/schedule at most NPRINT_QUEUE_MAX nprint calls
+            # (not all up front, in case this is VERY large, to spare RAM).
+            futures = {
+                executor.submit(self.execute_nprint, *args)
+                for args in itertools.islice(file_stream, NPRINT_QUEUE_MAX)
+            }
+
+            if self.args.verbosity >= 3:
+                print('Enqueued', len(futures), 'nPrint task(s)',
+                      'to be processed by at most', self.args.concurrency, 'worker(s)')
+
+            all_done = False
+            failure_count = 0
+
+            while futures:
+                # Wait only long enough for at least one call to complete.
+                (completed, futures) = concurrent.futures.wait(
+                    futures,
+                    return_when=concurrent.futures.FIRST_COMPLETED,
+                )
+
+                added_count = 0
+
+                for future in completed:
+                    # Top off the queue with at most one remaining call
+                    # for each completed call.
+                    for (added_count, args) in enumerate(itertools.islice(file_stream, 1),
+                                                         added_count + 1):
+                        futures.add(executor.submit(self.execute_nprint, *args))
+
+                    if not all_done and added_count == 0 and not futures:
+                        all_done = True
+
+                        if self.args.verbosity >= 3:
+                            print('nPrint tasks complete:', timing.time_elapsed)
+
+                    # Share the result.
+                    try:
+                        yield future.result()
+                    except execute.nprint.CommandError:
+                        failure_count += 1
+
+                if added_count and self.args.verbosity >= 3:
+                    print('Enqueued', added_count, 'more nPrint task(s)')
+
+            if failure_count > 0:
+                print('nPrint task(s) failed:', failure_count)
+
+    def __pre__(self, parser, args, results):
         try:
             warn_version_mismatch()
-        except nprint.NoCommand:
-            args.__parser__.error("nprint command could not be found on PATH "
-                                  "(to install see nprint-install)")
+        except execute.nprint.NoCommand:
+            parser.error("nprint command could not be found on PATH "
+                         "(to install see nprint-install)")
 
-        outdir = self.make_output_directory(args)
+        self.args = args
 
-        pcap_files = self.generate_files(args)
+    def __call__(self, args, results):
+        self.write_meta(results.meta)
 
-        processes = self.generate_procs(args, pcap_files, outdir)
+        self.ensure_output_directory()
 
-        self.pool_procs(processes, size=args.concurrency)
+        pcap_files = self.generate_pcaps()
 
-        return NetResult(outdir)
+        file_stream = self.filtermap_pcaps(pcap_files, results.labels)
+
+        # Time stream specially as it continues even after step formally completed:
+        stream_timing = results.__timing_steps__[self.generate_npts] = pipeline.Timing()
+
+        npt_stream = self.generate_npts(file_stream, stream_timing)
+
+        # Initialize results generator eagerly; (it will handle own buffer):
+        active_stream = prime_iterator(npt_stream)
+
+        return NetResult(
+            nprint_path=self.output_directory,
+            nprint_stream=active_stream,
+        )
 
 
 def warn_version_mismatch():
     """Warn if nPrint intended version doesn't match what's on PATH."""
-    version_result = nprint('--version', stdout=nprint.PIPE)
+    version_result = execute.nprint('--version', stdout=execute.nprint.PIPE)
     version_output = version_result.stdout.decode()
     version_match = re.match(r'nprint ([.\d]+)', version_output)
     if version_match:
@@ -337,10 +451,10 @@ class FileAccessType:
         return self.modes[self.access]
 
     def __call__(self, path):
-        if os.access(path, self.access):
+        if os.path.isfile(path) and os.access(path, self.access):
             return path
 
-        raise argparse.ArgumentTypeError(f"can't open '{path}' ({self.mode})")
+        raise argparse.ArgumentTypeError(f"can't access '{path}' ({self.mode})")
 
 
 class DirectoryAccessType:
