@@ -139,6 +139,8 @@ metaclass level, via `__default_registry__`.)
 """
 import abc
 import collections
+import itertools
+import time
 import types
 
 
@@ -188,6 +190,8 @@ class Step(metaclass=StepMeta):
     #: Sequences of keys, namedtuples and dataclasses are supported.
     __provides__ = ()
 
+    __pre_provides__ = ()
+
     #: Sequence of data keys on the results object -- gathered from the
     #: execution of preceding steps -- required for this step to execute
     __requires__ = ()
@@ -213,6 +217,14 @@ class Step(metaclass=StepMeta):
 
         """
 
+    def __pre__(self, parser, args, results):
+        """Execute preparatory actions.
+
+        Invoked before pipeline is run and Steps' main functionality
+        (`__call__`) is invoked.
+
+        """
+
     @abc.abstractmethod
     def __call__(self, args, results):
         """Step's executive functionality.
@@ -230,6 +242,10 @@ class Step(metaclass=StepMeta):
     def __repr__(self):
         return 'step:' + self.__class__.__name__
 
+    @property
+    def __name__(self):
+        return self.__class__.__name__.lower()
+
 
 class Pipeline(list):
     """Pipeline of executable Steps, instantiated from and extending an
@@ -296,8 +312,8 @@ class Pipeline(list):
         """
         # input registry may be any iterable
         steps = set(registry)
+        results = set(itertools.chain.from_iterable(step.__pre_provides__ for step in steps))
         run = []
-        results = set()
 
         while steps:
             steps_satisfied = [step for step in steps if results.issuperset(step.__requires__)]
@@ -331,26 +347,39 @@ class Pipeline(list):
 
         self.results = None
 
-    def __call__(self, args, results=None):
+    def __call__(self, parser, args, results=None, **kwargs):
         """Construct an iterator to execute the steps of the pipeline."""
         self.results = self.results_class() if results is None else results
+        self.results.__dict__.update(kwargs)
 
-        run = []
+        self.results.__timing__ = Timing()
+        self.results.__timing_steps__ = {}
 
-        for step in self[:]:
-            results_available = set(self.results.__dict__.keys())
+        with self.results.__timing__:
+            # pre-pipline
+            for step in self:
+                step.__pre__(parser, args, self.results)
 
-            if not results_available.issuperset(step.__requires__):
-                raise StepRequirementError(run, self, results_available)
+            run = []
 
-            step_results = step(args, self.results)
-            if step_results is not None:
-                self.merge(step_results)
+            for step in self[:]:
+                results_available = set(self.results.__dict__.keys())
 
-            run.append(step)
-            self.remove(step)
+                if not results_available.issuperset(step.__requires__):
+                    raise StepRequirementError(run, self, results_available)
 
-            yield (step, step_results)
+                step_timing = self.results.__timing_steps__[step] = Timing()
+
+                with step_timing:
+                    step_results = step(args, self.results)
+
+                if step_results is not None:
+                    self.merge(step_results)
+
+                run.append(step)
+                self.remove(step)
+
+                yield (step, step_results)
 
     def exhaust(self, args, results=None):
         """Execute the steps of the pipeline (non-iteratively)."""
@@ -392,3 +421,58 @@ class StepRequirementError(PipelineError):
         self.steps_run = steps_run
         self.steps_remaining = steps_remaining
         self.results_available = results_available
+
+
+class Timing:
+
+    class StaleTimer(ValueError):
+        pass
+
+    def __init__(self):
+        self.t0 = self.t1 = self.proc0 = self.proc1 = None
+
+    def start(self):
+        if self.t0 is not None:
+            raise self.StaleTimer
+
+        self.t0 = time.time()
+        self.proc0 = time.process_time()
+
+    def stop(self):
+        if self.t1 is not None:
+            raise self.StaleTimer
+
+        self.t1 = time.time()
+        self.proc1 = time.process_time()
+
+    def __enter__(self):
+        self.start()
+        return self
+
+    def __exit__(self, *exc_info):
+        self.stop()
+
+    @property
+    def time_elapsed(self):
+        if self.t0 is None or self.t1 is None:
+            return None
+
+        return self.t1 - self.t0
+
+    @property
+    def proc_elapsed(self):
+        if self.proc0 is None or self.proc1 is None:
+            return None
+
+        return self.proc1 - self.proc0
+
+    def __iter__(self):
+        yield self.time_elapsed
+        yield self.proc_elapsed
+
+    def __repr__(self):
+        return f'<{self.__class__.__name__}: {self}>'
+
+    def __str__(self):
+        return (f'total time elapsed {self.time_elapsed} | '
+                f'process time elapsed {self.proc_elapsed}')
